import assert from "assert";
import { filterNull } from "../../helpers";
function getTypeScriptTypeForField({
  fieldName,
  isRequired,
  isScalarType: isScalar,
  typeName
}) {
  if (isRequired) {
    if (isScalar) {
      return [fieldName, `Scalars["${typeName}"]`];
    }
    return [fieldName, typeName];
  }
  if (isScalar) {
    return [`${fieldName}?`, `Maybe<Scalars["${typeName}"]>`];
  }
  return [`${fieldName}?`, `Maybe<${typeName}>`];
}
function marshallField({
  columnName,
  fieldName,
  isDateType
}) {
  if (columnName === "ttl") {
    return `Math.floor(input.${fieldName}.getTime()/1000)`;
  }
  if (isDateType) {
    return `input.${fieldName} === null ? null : input.${fieldName}.toISOString()`;
  }
  return `input.${fieldName}`;
}
function makeKeyTemplate(prefix, fields, mode) {
  const accessors = [
    ...prefix ? [`'${prefix}'`] : [],
    ...fields.map((field) => {
      const { fieldName } = field;
      if (fieldName === "createdAt") {
        if (mode === "blind") {
          return "'createdAt' in input && typeof input.createdAt !== 'undefined' ? input.createdAt.getTime() : now.getTime()";
        }
        if (mode === "create") {
          return "now.getTime()";
        }
        if (mode === "read") {
          return "input.createdAt.getTime()";
        }
        assert.fail("Invalid mode");
      }
      if (fieldName === "updatedAt") {
        return "now.getTime()";
      }
      if (fieldName === "publicId") {
        if (mode === "create") {
          return "publicId";
        }
      }
      return marshallField(field);
    })
  ].filter(filterNull);
  return `[${accessors.join(", ")}].join('#')`;
}
function objectToString(obj) {
  return `{${Object.entries(obj).map(([k, value]) => `${k}: ${value}`)}}`;
}
function getTransformString(field) {
  if (field.columnName === "ttl") {
    return "(v) => new Date(v * 1000)";
  }
  if (field.isDateType) {
    return "(v) => new Date(v)";
  }
  return "";
}
function unmarshallFieldValue(field) {
  const transformString = getTransformString(field);
  const func = field.isRequired ? "unmarshallRequiredField" : "unmarshallOptionalField";
  const args = [
    "item",
    `'${field.fieldName}'`,
    `[${field.columnNamesForRead.map((c) => `'${c}'`).join(",")}]`,
    transformString
  ];
  return `${func}(${args.join(", ")})`;
}
function unmarshallField(field) {
  const out = unmarshallFieldValue(field);
  return `${field.fieldName}: ${out}`;
}
function handleCommonErrors() {
  return `
    if (err instanceof AssertionError || err instanceof BaseDataLibraryError) {
      throw err;
    }
    if (err instanceof ServiceException) {
      throw new UnexpectedAwsError(err);
    }
    throw new UnexpectedError(err);
  `;
}
function makeKeyForRead(config, key) {
  if (key.isComposite) {
    const doLegacy = config.legacyEmptySortFieldBehavior && key.sortKeyFields.length === 0;
    return {
      pk: makeKeyTemplate(
        key.partitionKeyPrefix,
        key.partitionKeyFields,
        "read"
      ),
      sk: doLegacy ? `'${key.sortKeyPrefix}#0'` : makeKeyTemplate(key.sortKeyPrefix, key.sortKeyFields, "read")
    };
  }
  return {
    pk: makeKeyTemplate(key.partitionKeyPrefix, key.partitionKeyFields, "read")
  };
}
export {
  getTypeScriptTypeForField,
  handleCommonErrors,
  makeKeyForRead,
  makeKeyTemplate,
  marshallField,
  objectToString,
  unmarshallField,
  unmarshallFieldValue
};
